@@ -17,6 +17,7 @@ from shutil import which
 from datetime import datetime
 from .tmpfile import tmpfile
 from .version import __version__
+from .completer import Completer
 from time import time
 import json
 
@@ -41,6 +42,25 @@ def prepJson(arg):
     return narg
 
 
+def _serGroup(g):
+    children = [_serGroup(child) for child in g.children]
+    return {"start":g.start, "end":g.end, "name":g.name, "children":children}
+
+def serGroup(g):
+    """
+    Serialize a Piraha group data structure (parse tree).
+    """
+    return {"text":g.text, "root":_serGroup(g)}
+
+def _deserGroup(data, text):
+    g = Group(data["name"], text, data["start"], data["end"])
+    g.children = [_deserGroup(c,text) for c in data["children"]]
+    return g
+
+def deserGroup(data):
+    return _deserGroup(data["root"],data["text"])
+
+
 # The way exit works is to raise SystemExit,
 # which may be caught. When we want to exit our
 # simulated shell, we should raise ShellExit
@@ -455,13 +475,51 @@ class shell:
         self.curr_ending = None
         self.last_pipe = None
         self.recursion = 0
-        self.max_recursion_depth = 20
+        self.max_recursion_depth = 2000
         log_file_dir = os.path.join(self.vars["HOME"],".evshell-logs")
         os.makedirs(log_file_dir, exist_ok = True)
         log_file = os.path.join(log_file_dir, f"log-{os.getpid()}.jtxt")
         self.log_fd = open(log_file,"w")
         self.log(msg="starting shell")
 
+    def serialize(self, fd):
+        print(json.dumps(self.vars),file=fd)
+        print(json.dumps(self.exports),file=fd)
+        print(json.dumps({
+            "max_recursion_depth":self.max_recursion_depth,
+            "cwd":os.getcwd(),
+        }),file=fd)
+
+        # Serialize shell functions
+        funcser = {}
+        for fname in self.funcs:
+            funcser[fname] = json.dumps([serGroup(x) for x in self.funcs[fname]])
+        print(json.dumps(funcser),file=fd)
+
+        import inspect
+        funcs = {}
+        for func in self.pyfuncs:
+            funcs[func] = inspect.getsource(self.pyfuncs[func])
+        print(json.dumps(funcs),file=fd)
+
+    def deserialize(self, fd):
+        self.vars = json.loads(fd.readline())
+        self.exports = json.loads(fd.readline())
+        data = json.loads(fd.readline())
+        self.max_recursion_depth = data["max_recursion_depth"]
+        os.chdir(data["cwd"])
+
+        # Deserialize shell functions
+        funcser = json.loads(fd.readline())
+        self.funcs = {}
+        for fname in funcser:
+            self.funcs[fname] = [deserGroup(x) for x in json.loads(funcser[fname])]
+
+        funcs = json.loads(fd.readline())
+        for func in funcs:
+            eval(func)
+            self.pyfuncs[func] = globals()[func]
+
     def err(self, e):
         print(colored(str(e),"red"),file=self.stderr)
 
@@ -1121,7 +1179,7 @@ class shell:
                 try:
                     return self.pyfuncs[args[0]](*args[1:])
                 except Exception as e:
-                    print(colored(f"'{args[0]}' threw '{type(e)}: {e}'"))
+                    print(colored(f"'{args[0]}' threw '{type(e)}: {e}'","red"))
                     return []
             elif args[0] == "unset":
                 for a in args[1:]:
@@ -1173,6 +1231,36 @@ class shell:
                             first_line = ""
                         if first_line.startswith("#!"):
                             args = re.split(r'\s+',first_line[2:].strip()) + args
+                if args[0] == 'storeenv':
+                    if len(args) != 2:
+                        print(f"Usage: storeenv 'name'")
+                    else:
+                        envfile = os.path.join(home, args[1])
+                        with open(envfile, "w") as fd:
+                            self.serialize(fd)
+                            print(f"Env stored to {envfile}")
+                    return []
+                elif args[0] == 'loadenv':
+                    if len(args) != 2:
+                        print(f"Usage: loadenv 'name'")
+                    else:
+                        envfile = os.path.join(home, args[1])
+                        with open(envfile, "r") as fd:
+                            self.deserialize(fd)
+                            print(f"Env loaded from {envfile}")
+                    return []
+                elif len(args) == 4 and args[0] == 'pyfrom' and args[2] == 'import':
+                    args = self.allow_cmd(args)
+                    modname = args[1]
+                    funcname = args[3]
+                    module = __import__(modname)
+                    self.pyfuncs[funcname] = getattr(module,funcname)
+                    return []
+                elif args[0] == 'exec':
+                    exec_cmd = which(args[1])
+                    args = self.allow_cmd(args[1:])
+                    if exec_cmd is not None:
+                        os.execve(exec_cmd,args,self.exports)
                 if not os.path.exists(args[0]):
                     if gr is None:
                         fno = 0
@@ -1313,6 +1401,9 @@ class shell:
 def interactive(shell):
     try:
         import readline
+        c = Completer()
+        readline.set_completer(c.complete)
+        readline.parse_and_bind("tab: complete")
     except Exception as ee:
         print(colored("Import of readline failed","red"))
     msg = "EVAL"
